@@ -11,7 +11,7 @@
 #             Lorenzo Bivens (pkgin package)          <lorenzobivens@gmail.com>
 # -----------------------------------------------------------------------------
 # Creation  : 26-Apr-2010
-# Last mod  : 05-Sep-2013
+# Last mod  : 28-Oct-2013
 # -----------------------------------------------------------------------------
 
 """
@@ -46,17 +46,28 @@ import base64, hashlib, os, re, string, tempfile, subprocess, types
 import tempfile, functools, StringIO
 import fabric, fabric.api, fabric.operations, fabric.context_managers, fabric.state
 
-VERSION               = "0.6.5"
-RE_SPACES             = re.compile("[\s\t]+")
-MAC_EOL               = "\n"
-UNIX_EOL              = "\n"
-WINDOWS_EOL           = "\r\n"
-MODE_LOCAL            = "CUISINE_MODE_LOCAL"
-MODE_SUDO             = "CUISINE_MODE_SUDO"
-SUDO_PASSWORD         = "CUISINE_SUDO_PASSWORD"
-OPTION_PACKAGE        = "CUISINE_OPTION_PACKAGE"
-OPTION_PYTHON_PACKAGE = "CUISINE_OPTION_PYTHON_PACKAGE"
-CMD_APT_GET           = 'DEBIAN_FRONTEND=noninteractive apt-get -q --yes -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confold" '
+try:
+	import reporter
+	reporter.StdoutReporter.Install()
+	logging = reporter.bind("cuisine")
+except ImportError:
+	import logging
+
+VERSION                 = "0.7.1"
+RE_SPACES               = re.compile("[\s\t]+")
+STRINGIFY_MAXSTRING     = 80
+STRINGIFY_MAXLISTSTRING = 20
+MAC_EOL                 = "\n"
+UNIX_EOL                = "\n"
+WINDOWS_EOL             = "\r\n"
+MODE_LOCAL              = "CUISINE_MODE_LOCAL"
+MODE_SUDO               = "CUISINE_MODE_SUDO"
+SUDO_PASSWORD           = "CUISINE_SUDO_PASSWORD"
+OPTION_PACKAGE          = "CUISINE_OPTION_PACKAGE"
+OPTION_PYTHON_PACKAGE   = "CUISINE_OPTION_PYTHON_PACKAGE"
+CMD_APT_GET             = 'DEBIAN_FRONTEND=noninteractive apt-get -q --yes -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confold" '
+SHELL_ESCAPE            = " '\";`|"
+STATS                   = None
 
 AVAILABLE_OPTIONS = dict(
 	package=["apt", "yum", "zypper", "pacman", "emerge", "pkgin"],
@@ -68,6 +79,124 @@ DEFAULT_OPTIONS = dict(
 	python_package="pip"
 )
 
+# =============================================================================
+#
+# STATS
+#
+# =============================================================================
+
+class Stats(object):
+	"""A work-in-progress class to store cuisine's statistics, so that you
+	can have a summary of what has been done."""
+
+	def __init__( self ):
+		self.filesRead         = []
+		self.filesWritten      = []
+		self.packagesInstalled = []
+
+# =============================================================================
+#
+# DECORATORS
+#
+# =============================================================================
+
+def stringify( value ):
+	"""Turns the given value in a user-friendly string that can be displayed"""
+	if   type(value) in (str, unicode, bytes) and len(value) > STRINGIFY_MAXSTRING:
+		return "{0}...".format(value[0:STRINGIFY_MAXSTRING])
+	elif type(value) in (list, tuple) and len(value) > 10:
+		return"[{0},...]".format(", ".join([stringify(_) for _ in value[0:STRINGIFY_MAXLISTSTRING]]))
+	else:
+		return str(value)
+
+def log_message( message ):
+	"""Logs the given message"""
+	logging.info( message )
+
+def log_call( function, args, kwargs ):
+	"""Logs the given function call"""
+	function_name = function.__name__
+	a = ", ".join([stringify(_) for _ in args] + [str(k) + "=" + stringify(v) for k,v in kwargs.items()])
+	log_message("{0}({1})".format(function_name, a))
+
+def logged(message=None):
+	"""Logs the invoked function name and arguments."""
+	# TODO: Options - prevent sub @logged to output anything
+	# TODO: Message - allow to specify a message
+	# TODO: Category - read/write/exec as well as mode
+	# [2013-10-28T10:18:32] user@host [sudo|user] [R/W] cuinine.function(xx,xxx,xx) [time]
+	# [2013-10-28T10:18:32] user@host [sudo|user] [!] Exception
+	def logged_wrapper(function, message=message):
+		def wrapper(*args, **kwargs):
+			log_call(function, args, kwargs)
+			return function(*args, **kwargs)
+		# We copy name and docstring
+		functools.update_wrapper(wrapper, function)
+		return wrapper
+	if type(message) == types.FunctionType:
+		return logged_wrapper(message, None)
+	else:
+		return logged_wrapper
+
+def dispatch(prefix=None):
+	"""Dispatches the current function to specific implementation. The `prefix`
+	parameter indicates the common option prefix, and the `select_[option]()`
+	function will determine the function suffix.
+
+	For instance the package functions are defined like this:
+
+	{{{
+	@dispatch("package")
+	def package_ensure(...):
+		...
+	def package_ensure_apt(...):
+		...
+	def package_ensure_yum(...):
+		...
+	}}}
+
+	and then when a user does
+
+	{{{
+	cuisine.select_package("yum")
+	cuisine.package_ensure(...)
+	}}}
+
+	then the `dispatch` function will dispatch `package_ensure` to
+	`package_ensure_yum`.
+
+	If your prefix is the first word of the function name before the
+	first `_` then you can simply use `@dispatch` without parameters.
+	"""
+	def dispatch_wrapper(function, prefix=prefix):
+		def wrapper(*args, **kwargs):
+			function_name = function.__name__
+			_prefix       = prefix or function_name.split("_")[0].replace(".","_")
+			select        = fabric.api.env.get("CUISINE_OPTION_" + _prefix.upper())
+			assert select, "No option defined for: %s, call select_%s(<YOUR OPTION>) to set it" % (_prefix.upper(), prefix.lower().replace(".","_"))
+			function_name = function.__name__ + "_" + select
+			specific      = eval(function_name)
+			if specific:
+				if type(specific) == types.FunctionType:
+					return specific(*args, **kwargs)
+				else:
+					raise Exception("Function expected for: " + function_name)
+			else:
+				raise Exception("Function variant not defined: " + function_name)
+		# We copy name and docstring
+		functools.update_wrapper(wrapper, function)
+		return wrapper
+	if type(prefix) == types.FunctionType:
+		return dispatch_wrapper(prefix, None)
+	else:
+		return dispatch_wrapper
+
+# =============================================================================
+#
+# MODES
+#
+# =============================================================================
+
 def sudo_password(password=None):
 	"""Sets the password for the sudo command."""
 	if password is None:
@@ -128,6 +257,10 @@ def is_local():  return mode(MODE_LOCAL)
 def is_remote(): return not mode(MODE_LOCAL)
 def is_sudo():   return mode(MODE_SUDO)
 
+def shell_safe( path ):
+	"""Makes sure that the given path/string is escaped and safe for shell"""
+	return "".join([("\\" + _) if _ in SHELL_ESCAPE else _ for _ in path])
+
 # =============================================================================
 #
 # OPTIONS
@@ -203,6 +336,7 @@ def cd(*args, **kwargs):
 	return fabric.api.cd(*args, **kwargs)
 
 
+@logged
 def sudo(*args, **kwargs):
 	"""A wrapper to Fabric's run/sudo commands, using the
 	'cuisine.MODE_SUDO' global to tell whether the command should be run as
@@ -210,6 +344,7 @@ def sudo(*args, **kwargs):
 	with mode_sudo():
 		return run(*args, **kwargs)
 
+@logged
 def connect( host, user="root"):
 	"""Sets Fabric's current host to the given host. This is useful when
 	using Cuisine in standalone."""
@@ -217,65 +352,6 @@ def connect( host, user="root"):
 	fabric.api.env.host_string = host
 	fabric.api.env.user        = user
 
-# =============================================================================
-#
-# DECORATORS
-#
-# =============================================================================
-
-def dispatch(prefix=None):
-	"""Dispatches the current function to specific implementation. The `prefix`
-	parameter indicates the common option prefix, and the `select_[option]()`
-	function will determine the function suffix.
-
-	For instance the package functions are defined like this:
-
-	{{{
-	@dispatch("package")
-	def package_ensure(...):
-		...
-	def package_ensure_apt(...):
-		...
-	def package_ensure_yum(...):
-		...
-	}}}
-
-	and then when a user does
-
-	{{{
-	cuisine.select_package("yum")
-	cuisine.package_ensure(...)
-	}}}
-
-	then the `dispatch` function will dispatch `package_ensure` to
-	`package_ensure_yum`.
-
-	If your prefix is the first word of the function name before the
-	first `_` then you can simply use `@dispatch` without parameters.
-	"""
-	def dispatch_wrapper(function, prefix=prefix):
-		def wrapper(*args, **kwargs):
-			function_name = function.__name__
-			_prefix       = prefix or function_name.split("_")[0].replace(".","_")
-			select        = fabric.api.env.get("CUISINE_OPTION_" + _prefix.upper())
-			assert select, "No option defined for: %s, call select_%s(<YOUR OPTION>) to set it" % (_prefix.upper(), prefix.lower().replace(".","_"))
-			function_name = function.__name__ + "_" + select
-			specific      = eval(function_name)
-			if specific:
-				if type(specific) == types.FunctionType:
-					return specific(*args, **kwargs)
-				else:
-					raise Exception("Function expected for: " + function_name)
-			else:
-				raise Exception("Function variant not defined: " + function_name)
-		# We copy name and docstring
-		functools.update_wrapper(wrapper, function)
-		return wrapper
-	if type(prefix) == types.FunctionType:
-		return dispatch_wrapper(prefix, None)
-	else:
-		return dispatch_wrapper
-
 # =============================================================================
 #
 # TEXT PROCESSING
@@ -310,7 +386,18 @@ def text_nospace(text):
 
 def text_replace_line(text, old, new, find=lambda old, new: old == new, process=lambda _: _):
 	"""Replaces lines equal to 'old' with 'new', returning the new
-	text and the count of replacements."""
+	text and the count of replacements.
+
+	Returns: (text, number of lines replaced)
+
+	`process` is a function that will pre-process each line (you can think of
+	it as a normalization function, by default it will return the string as-is),
+	and `find` is the function that will compare the current line to the
+	`old` line.
+
+	The finds the line using `find(process(current_line), process(old_line))`,
+	and if this matches, will insert the new line instead.
+	"""
 	res = []
 	replaced = 0
 	eol = text_detect_eol(text)
@@ -342,6 +429,19 @@ def text_ensure_line(text, *lines):
 	return eol.join(res)
 
 def text_strip_margin(text, margin="|"):
+	"""Will strip all the characters before the left margin identified
+	by the `margin` character in your text. For instance
+
+	```
+			|Hello, world!
+	```
+
+	will result in
+
+	```
+	Hello, world!
+	```
+	"""
 	res = []
 	eol = text_detect_eol(text)
 	for line in text.split(eol):
@@ -363,6 +463,7 @@ def text_template(text, variables):
 #
 # =============================================================================
 
+@logged
 def file_local_read(location):
 	"""Reads a *local* file from the given location, expanding '~' and
 	shell variables."""
@@ -372,6 +473,22 @@ def file_local_read(location):
 	f.close()
 	return t
 
+
+@logged
+def file_backup(location, suffix=".orig", once=False):
+	"""Backups the file at the given location in the same directory, appending
+	the given suffix. If `once` is True, then the backup will be skipped if
+	there is already a backup file."""
+	backup_location = location + suffix
+	if once and file_exists(backup_location):
+		return False
+	else:
+		return run("cp -a {0} {1}".format(
+			shell_safe(location),
+			shell_safe(backup_location)
+		))
+
+@logged
 def file_read(location, default=None):
 	"""Reads the *remote* file at the given location, if default is not `None`,
 	default will be returned if the file does not exist."""
@@ -383,39 +500,42 @@ def file_read(location, default=None):
 	with fabric.context_managers.settings(
 		fabric.api.hide('stdout')
 	):
-		frame = run('cat "%s" | openssl base64' % (location))
+		frame = run("cat {0} | openssl base64".format(shell_safe((location))))
 		return base64.b64decode(frame)
 
 def file_exists(location):
 	"""Tests if there is a *remote* file at the given location."""
-	return run('test -e "%s" && echo OK ; true' % (location)).endswith("OK")
+	return run('test -e %s && echo OK ; true' % (shell_safe(location))).endswith("OK")
 
 def file_is_file(location):
-	return run("test -f '%s' && echo OK ; true" % (location)).endswith("OK")
+	return run("test -f %s && echo OK ; true" % (shell_safe(location))).endswith("OK")
 
 def file_is_dir(location):
-	return run("test -d '%s' && echo OK ; true" % (location)).endswith("OK")
+	return run("test -d %s && echo OK ; true" % (shell_safe(location))).endswith("OK")
 
 def file_is_link(location):
-	return run("test -L '%s' && echo OK ; true" % (location)).endswith("OK")
+	return run("test -L %s && echo OK ; true" % (shell_safe(location))).endswith("OK")
 
+@logged
 def file_attribs(location, mode=None, owner=None, group=None):
 	"""Updates the mode/owner/group for the remote file at the given
 	location."""
 	return dir_attribs(location, mode, owner, group, False)
 
+@logged
 def file_attribs_get(location):
 	"""Return mode, owner, and group for remote path.
 	Return mode, owner, and group if remote path exists, 'None'
 	otherwise.
 	"""
 	if file_exists(location):
-		fs_check = run('stat %s %s' % (location, '--format="%a %U %G"'))
+		fs_check = run('stat %s %s' % (shell_safe(location), '--format="%a %U %G"'))
 		(mode, owner, group) = fs_check.split(' ')
 		return {'mode': mode, 'owner': owner, 'group': group}
 	else:
 		return None
 
+@logged
 def file_write(location, content, mode=None, owner=None, group=None, sudo=None, check=True, scp=False):
 	"""Writes the given content to the file at the given remote
 	location, optionally setting mode/owner/group."""
@@ -429,11 +549,11 @@ def file_write(location, content, mode=None, owner=None, group=None, sudo=None,
 	if sig != file_md5(location):
 		if is_local():
 			with mode_sudo(use_sudo):
-				run('cp "%s" "%s"'%(local_path,location))
+				run('cp %s %s'%(shell_safe(local_path), shell_safe(location)))
 		else:
 			if scp:
 				hostname = fabric.api.env.host_string if len(fabric.api.env.host_string.split(':')) == 1 else fabric.api.env.host_string.split(':')[0]
-				scp_cmd = 'scp "%s" "%s"@"%s":"%s"'%(local_path,fabric.api.env.user,hostname,location)
+				scp_cmd = 'scp %s %s@%s:%s'% (shell_safe(local_path), shell_safe(fabric.api.env.user), shell_safe(hostname), shell_safe(location))
 				print('[localhost] ' +  scp_cmd)
 				run_local(scp_cmd)
 			else:
@@ -447,7 +567,7 @@ def file_write(location, content, mode=None, owner=None, group=None, sudo=None,
 					**{MODE_SUDO: use_sudo}
 				):
 					# See: http://unix.stackexchange.com/questions/22834/how-to-uncompress-zlib-data-in-unix
-					result = run("echo '%s' | openssl base64 -A -d -out \"%s\"" % (base64.b64encode(content), location))
+					result = run("echo '%s' | openssl base64 -A -d -out %s" % (base64.b64encode(content), shell_safe(location)))
 					if "openssl:Error" in result:
 						fabric.api.abort('cuisine.file_write("%s",...) failed because openssl does not support base64 command.' % (location))
 	# Remove the local temp file
@@ -462,6 +582,7 @@ def file_write(location, content, mode=None, owner=None, group=None, sudo=None,
 	with mode_sudo(use_sudo):
 		file_attribs(location, mode=mode, owner=owner, group=group)
 
+@logged
 def file_ensure(location, mode=None, owner=None, group=None, scp=False):
 	"""Updates the mode/owner/group for the remote file at the given
 	location."""
@@ -470,6 +591,7 @@ def file_ensure(location, mode=None, owner=None, group=None, scp=False):
 	else:
 		file_write(location,"",mode=mode,owner=owner,group=group,scp=scp)
 
+@logged
 def file_upload(remote, local, sudo=None, scp=False):
 	"""Uploads the local file to the remote location only if the remote location does not
 	exists or the content are different."""
@@ -482,43 +604,55 @@ def file_upload(remote, local, sudo=None, scp=False):
 	if not file_exists(remote) or sig != file_md5(remote):
 		if is_local():
 			if use_sudo:
-				globals()['sudo']('cp "%s" "%s"'%(local,remote))
+				globals()['sudo']('cp %s %s'%(shell_safe(local), shell_safe(remote)))
 			else:
 				run('cp "%s" "%s"'%(local,remote))
 		else:
 			if scp:
 				hostname = fabric.api.env.host_string if len(fabric.api.env.host_string.split(':')) == 1 else fabric.api.env.host_string.split(':')[0]
-				scp_cmd = 'scp "%s" "%s"@"%s":"%s"'%(local,fabric.api.env.user,hostname,remote)
+				scp_cmd = 'scp %s %s@%s:%s'%( shell_safe(local), shell_safe(fabric.api.env.user), shell_safe(hostname), shell_safe(remote))
 				print('[localhost] ' +  scp_cmd)
 				run_local(scp_cmd)
 			else:
 				fabric.operations.put(local, remote, use_sudo=use_sudo)
 
+@logged
 def file_update(location, updater=lambda x: x):
 	"""Updates the content of the given by passing the existing
 	content of the remote file at the given location to the 'updater'
-	function.
+	function. Return true if file content was changed.
 
 	For instance, if you'd like to convert an existing file to all
 	uppercase, simply do:
 
 	>   file_update("/etc/myfile", lambda _:_.upper())
+
+	Or restart service on config change:
+
+	>   if file_update("/etc/myfile.cfg", lambda _: text_ensure_line(_, line)): run("service restart")
 	"""
 	assert file_exists(location), "File does not exists: " + location
-	new_content = updater(file_read(location))
+	old_content = file_read(location)
+	new_content = updater(old_content)
+	if (old_content == new_content):
+		return False
 	# assert type(new_content) in (str, unicode, fabric.operations._AttributeString), "Updater must be like (string)->string, got: %s() = %s" %  (updater, type(new_content))
-	run('echo "%s" | openssl base64 -A -d -out "%s"' % (base64.b64encode(new_content), location))
+	file_write(location, new_content)
+	return True
 
+@logged
 def file_append(location, content, mode=None, owner=None, group=None):
 	"""Appends the given content to the remote file at the given
 	location, optionally updating its mode/owner/group."""
-	run('echo "%s" | openssl base64 -A -d >> "%s"' % (base64.b64encode(content), location))
+	run('echo "%s" | openssl base64 -A -d >> %s' % (base64.b64encode(content), shell_safe(location)))
 	file_attribs(location, mode, owner, group)
 
+@logged
 def file_unlink(path):
 	if file_exists(path):
-		run("unlink '%s'" % (path))
+		run("unlink %s" % (shell_safe(path)))
 
+@logged
 def file_link(source, destination, symbolic=True, mode=None, owner=None, group=None):
 	"""Creates a (symbolic) link between source and destination on the remote host,
 	optionally setting its mode/owner/group."""
@@ -528,25 +662,27 @@ def file_link(source, destination, symbolic=True, mode=None, owner=None, group=N
 	if file_is_link(destination):
 		file_unlink(destination)
 	if symbolic:
-		run('ln -sf "%s" "%s"' % (source, destination))
+		run('ln -sf %s %s' % (shell_safe(source), shell_safe(destination)))
 	else:
-		run('ln -f "%s" "%s"' % (source, destination))
+		run('ln -f %s %s' % (shell_safe(source), shell_safe(destination)))
 	file_attribs(destination, mode, owner, group)
 
+@logged
 def file_sha256(location):
 	"""Returns the SHA-256 sum (as a hex string) for the remote file at the given location."""
 	# NOTE: In some cases, sudo can output errors in here -- but the errors will
 	# appear before the result, so we simply split and get the last line to
 	# be on the safe side.
-	sig = run('shasum -a 256 "%s" | cut -d" " -f1' % (location)).split("\n")
+	sig = run('shasum -a 256 %s | cut -d" " -f1' % (shell_safe(location))).split("\n")
 	return sig[-1].strip()
 
+@logged
 def file_md5(location):
 	"""Returns the MD5 sum (as a hex string) for the remote file at the given location."""
 	# NOTE: In some cases, sudo can output errors in here -- but the errors will
 	# appear before the result, so we simply split and get the last line to
 	# be on the safe side.
-	sig = run('md5sum "%s" | cut -d" " -f1' % (location)).split("\n")
+	sig = run('md5sum %s | cut -d" " -f1' % (shell_safe(location))).split("\n")
 	return sig[-1].strip()
 
 # =============================================================================
@@ -555,6 +691,7 @@ def file_md5(location):
 #
 # =============================================================================
 
+@logged
 def process_find(name, exact=False):
 	"""Returns the pids of processes with the given name. If exact is `False`
 	it will return the list of all processes that start with the given
@@ -582,6 +719,7 @@ def process_find(name, exact=False):
 			res.append(pid)
 	return res
 
+@logged
 def process_kill(name, signal=9, exact=False):
 	"""Kills the given processes with the given name. If exact is `False`
 	it will return the list of all processes that start with the given
@@ -595,27 +733,29 @@ def process_kill(name, signal=9, exact=False):
 #
 # =============================================================================
 
+@logged
 def dir_attribs(location, mode=None, owner=None, group=None, recursive=False):
 	"""Updates the mode/owner/group for the given remote directory."""
 	recursive = recursive and "-R " or ""
 	if mode:
-		run('chmod %s %s "%s"' % (recursive, mode,  location))
+		run('chmod %s %s %s' % (recursive, mode,  shell_safe(location)))
 	if owner:
-		run('chown %s %s "%s"' % (recursive, owner, location))
+		run('chown %s %s %s' % (recursive, owner, shell_safe(location)))
 	if group:
-		run('chgrp %s %s "%s"' % (recursive, group, location))
+		run('chgrp %s %s %s' % (recursive, group, shell_safe(location)))
 
 def dir_exists(location):
 	"""Tells if there is a remote directory at the given location."""
-	return run('test -d "%s" && echo OK ; true' % (location)).endswith("OK")
+	return run('test -d %s && echo OK ; true' % (shell_safe(location))).endswith("OK")
 
+@logged
 def dir_remove(location, recursive=True):
 	""" Removes a directory """
 	flag = ''
 	if recursive:
 		flag = 'r'
 	if dir_exists(location):
-		return run('rm -%sf %s && echo OK ; true' % (flag, location))
+		return run('rm -%sf %s && echo OK ; true' % (flag, shell_safe(location)))
 
 def dir_ensure(location, recursive=False, mode=None, owner=None, group=None):
 	"""Ensures that there is a remote directory at the given location,
@@ -624,7 +764,7 @@ def dir_ensure(location, recursive=False, mode=None, owner=None, group=None):
 	If we are not updating the owner/group then this can be done as a single
 	ssh call, so use that method, otherwise set owner/group after creation."""
 	if not dir_exists(location):
-		run('mkdir %s "%s"' % (recursive and "-p" or "", location))
+		run('mkdir %s %s' % (recursive and "-p" or "", shell_safe(location)))
 	if owner or group or mode:
 		dir_attribs(location, owner=owner, group=group, mode=mode, recursive=recursive)
 
@@ -634,30 +774,36 @@ def dir_ensure(location, recursive=False, mode=None, owner=None, group=None):
 #
 # =============================================================================
 
+@logged
 @dispatch
 def package_upgrade(distupgrade=False):
 	"""Updates every package present on the system."""
 
+@logged
 @dispatch
 def package_update(package=None):
 	"""Updates the package database (when no argument) or update the package
 	or list of packages given as argument."""
 
+@logged
 @dispatch
 def package_install(package, update=False):
 	"""Installs the given package/list of package, optionally updating
 	the package database."""
 
+@logged
 @dispatch
 def package_ensure(package, update=False):
 	"""Tests if the given package is installed, and installs it in
 	case it's not already there. If `update` is true, then the
 	package will be updated if it already exists."""
 
+@logged
 @dispatch
 def package_clean(package=None):
 	"""Clean the repository for un-needed files."""
 
+@logged
 @dispatch
 def package_remove(package, autoclean=False):
 	"""Remove package and optionally clean unused packages"""
@@ -666,6 +812,7 @@ def package_remove(package, autoclean=False):
 # APT PACKAGE (DEBIAN/UBUNTU)
 # -----------------------------------------------------------------------------
 
+@logged
 def repository_ensure_apt(repository):
 	package_ensure_apt('python-software-properties')
 	sudo("add-apt-repository --yes " + repository)
@@ -1102,7 +1249,7 @@ def user_passwd(name, passwd, encrypted_passwd=True):
 		sudo("usermod -p '%s' %s" % (passwd,name))
 	else:
 		# NOTE: We use base64 here in case the password contains special chars
-		sudo("echo %s | openssl base64 -A -d | chpasswd" % (encoded_password))
+		sudo("echo %s | openssl base64 -A -d | chpasswd" % (shell_safe(encoded_password)))
 
 def user_create(name, passwd=None, home=None, uid=None, gid=None, shell=None,
 	uid_min=None, uid_max=None, encrypted_passwd=True, fullname=None, createhome=True):
@@ -1360,15 +1507,39 @@ def upstart_ensure(name):
 	with fabric.api.settings(warn_only=True):
 		status = sudo("service %s status" % name)
 	if status.failed:
-		sudo("service %s start" % name)
+		status = sudo("service %s start" % name)
+	return status
 
+def upstart_reload(name):
+	"""Reloads the given service, or starts it if it is not running."""
+	with fabric.api.settings(warn_only=True):
+		status = sudo("service %s reload" % name)
+	if status.failed:
+		status = sudo("service %s start" % name)
+	return status
+
+def upstart_restart(name):
+	"""Tries a `restart` command to the given service, if not successful
+	will stop it and start it. If the service is not started, will start it."""
+	with fabric.api.settings(warn_only=True):
+		status = sudo("service %s status" % name)
+	if status.failed:
+		return sudo("service %s start" % name)
+	else:
+		status = sudo("service %s restart" % name)
+		if status.failed:
+			sudo("service %s stop"  % name)
+			return sudo("service %s start" % name)
+		else:
+			return status
 
 def upstart_stop(name):
 	"""Ensures that the given upstart service is stopped."""
 	with fabric.api.settings(warn_only=True):
 		status = sudo("service %s status" % name)
 	if status.succeeded:
-		sudo("service %s stop" % name)
+		status = sudo("service %s stop" % name)
+	return status
 
 
 # =============================================================================
@@ -1397,7 +1568,6 @@ def system_uuid():
 #
 # =============================================================================
 
-
 def locale_check(locale):
 	locale_data = sudo("locale -a | egrep '^%s$' ; true" % (locale,))
 	return locale_data == locale
@@ -1410,6 +1580,7 @@ def locale_ensure(locale):
 
 # Sets up the default options so that @dispatch'ed functions work
 def _init():
+	STATS = Stats()
 	for option, value in DEFAULT_OPTIONS.items():
 		eval("select_" + option)(value)
 
